import dataclasses
from enum import Enum
from collections.abc import Callable, Coroutine
from typing import Any, Generic, TypeAlias, TypeVar

from pydantic.generics import GenericModel


_T = TypeVar("_T")
_S = TypeVar("_S")

ValidatorFunc: TypeAlias = Callable[[Any], Coroutine[Any, Any, None]]


class ErrorCodeEnum(str, Enum):
    unique_constraint = "UNIQUE CONSTRAINT"
    not_found = "NOT FOUND"
    permission_denied = "PERMISSION DENIED"


class ErrorSchema(GenericModel, Generic[_S]):
    code: str
    message: str
    detail: str | None = None
    source: _S | None = None


class ValidationError(Generic[_T], Exception):
    messages: list[_T]

    def __init__(self, messages: list[_T], *args: object) -> None:
        super().__init__(*args)
        self.messages = messages


@dataclasses.dataclass()
class ValidationContext(Generic[_T]):
    _errors: list[_T] = dataclasses.field(default_factory=list)

    def add_error(self, error: _T) -> None:
        self._errors.append(error)

    def extend_nested_error(self, errors: list[_T]) -> None:
        self._errors.extend(errors)

    @property
    def errors(self) -> list[_T]:
        return self._errors



from .validator import Validator, validate
from .types import ErrorSchema, ValidationError, ErrorCodeEnum, ValidationContext

__all__ = [
    "Validator",
    "validate",
    "ErrorSchema",
    "ErrorCodeEnum",
    "ValidationError",
    "ValidationContext",
]



import functools
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import Any, Generic
from business_validator.types import (
    ValidationContext,
    ValidationError,
    ValidatorFunc,
    _T,
)


def validate(name: str | None = None) -> Callable[[ValidatorFunc], ValidatorFunc]:
    def decorator(f: ValidatorFunc) -> ValidatorFunc:
        @wraps(f)
        async def wrapper(self: Validator[Any]) -> None:
            await f(self)

        wrapper.name = name  # type: ignore[attr-defined]
        wrapper.__is_validator__ = True  # type: ignore[attr-defined]

        return wrapper

    return decorator


class Validator(Generic[_T]):
    async def validate(self) -> None:
        context = self.context

        for validator in self.get_validation_methods():
            await validator(self)

        if context.errors:
            raise ValidationError(context.errors)

    @functools.cached_property
    def context(self) -> ValidationContext[_T]:
        return ValidationContext()

    async def errors(self) -> list[_T]:
        try:
            await self.validate()
        except ValidationError as e:
            return e.messages
        return []

    @classmethod
    def get_validation_methods(
        cls,
    ) -> list[Callable[["Validator[_T]"], Awaitable[None]]]:
        return [
            value
            for value in cls.__dict__.values()
            if getattr(value, "__is_validator__", False)
        ]



import asyncio
import dataclasses
from pydantic import BaseModel
from business_validator import (
    Validator,
    ValidationError,
    ErrorSchema,
    ErrorCodeEnum,
    validate,
)


class CommentDto(BaseModel):
    comment: str
    post_id: int
    owner_id: int


class Source(BaseModel):
    local: str


@dataclasses.dataclass()
class CommentValidator(Validator[ErrorSchema[Source]]):
    dto: CommentDto

    @validate()
    async def test1(self):
        post_ids = list(range(1, 10))

        if self.dto.post_id not in post_ids:
            self.context.add_error(
                ErrorSchema(
                    code=ErrorCodeEnum.not_found.value,
                    message="Id doen't not exists",
                    detail=f"Post with id={self.dto.post_id} not found",
                    source=Source(
                        local="data/post_id",
                    ),
                )
            )

    @validate()
    async def test2(self):
        owner_ids = list(range(1, 10))

        if self.dto.owner_id not in owner_ids:
            self.context.add_error(
                ErrorSchema(
                    code=ErrorCodeEnum.not_found.value,
                    message="Id doen't not exists",
                    detail=f"User with id={self.dto.post_id} not found",
                    source=Source(
                        local="data/owner_id",
                    ),
                )
            )


async def main():
    dto = CommentDto(comment="comment", owner_id=1234, post_id=222)

    try:
        await CommentValidator(dto).validate()
    except ValidationError as e:
        print(e.messages)


if __name__ == "__main__":
    asyncio.run(main())
